/******************************************************************************/
/*  Econonavit data reciever                                                  */
/*  Main                                                                      */
/*  Copyright(c) 2013 Hirofumi Momose  hirofumi[at]momose.com                 */
/******************************************************************************/

/** INCLUDES *******************************************************/
#include <xc.h>
#include "Compiler.h"
#include "USB/usb.h"
#include "usb_function_cdc.h"

#ifndef MIN
#define MIN(a,b)	((a) < (b) ? (a) : (b))
#endif

#ifndef MAX
#define MAX(a,b)	((a) > (b) ? (a) : (b))
#endif

/** CONFIGURATION **************************************************/
    #pragma config PLLDIV   = 1         // Use 4 MHz crystal
    /* Choose one of followings
     * 1 No prescale (4 MHz oscillator input drives PLL directly)
     * 2 Divide by 2 (8 MHz oscillator input)
     * 3 Divide by 3 (12 MHz oscillator input)
     * 4 Divide by 4 (16 MHz oscillator input)
     * 5 Divide by 5 (20 MHz oscillator input)
     * 6 Divide by 6 (24 MHz oscillator input)
     * 10 Divide by 10 (40 MHz oscillator input)
     * 12 Divide by 12 (48 MHz oscillator input)
     */

    #pragma config CPUDIV   = OSC4_PLL6
    /* Choose one of followings
     * OSC1_PLL2 [Primary Oscillator Src: /1][96 MHz PLL Src: /2]
     * OSC2_PLL3 [Primary Oscillator Src: /2][96 MHz PLL Src: /3]
     * OSC3_PLL4 [Primary Oscillator Src: /3][96 MHz PLL Src: /4]
     * OSC4_PLL6 [Primary Oscillator Src: /4][96 MHz PLL Src: /6]
     */

    #pragma config USBDIV   = 2         // Clock source from 96MHz PLL/2

    #pragma config FOSC     = HSPLL_HS
    /* Choose one of followings
     * INTOSC_EC Internal oscillator, CLKO function on RA6, EC used by USB (INTCKO)
     * ECPLLIO_EC EC oscillator, PLL enabled, port function on RA6 (ECPIO)
     * XT_XT XT oscillator (XT)
     * INTOSC_HS Internal oscillator, HS oscillator used by USB (INTHS)
     * ECPLL_EC EC oscillator, PLL enabled, CLKO function on RA6 (ECPLL)
     * INTOSC_XT Internal oscillator, XT used by USB (INTXT)
     * EC_EC EC oscillator, CLKO function on RA6 (EC)
     * ECIO_EC EC oscillator, port function on RA6 (ECIO)
     * XTPLL_XT XT oscillator, PLL enabled (XTPLL)
     * HSPLL_HS HS oscillator, PLL enabled (HSPLL)
     * HS HS oscillator (HS)
     */

    #pragma config FCMEN    = OFF       // Fail safe clock disabled
    #pragma config IESO     = OFF       // Int / Ext clock switchover
    #pragma config PWRT     = OFF       // Power up timer
    #pragma config BOR      = OFF       // Brown out reset
    #pragma config BORV     = 3
    #pragma config VREGEN   = OFF       //USB Voltage Regulator
    #pragma config WDT      = OFF
    #pragma config WDTPS    = 32768
    #pragma config CCP2MX   = ON
    #pragma config PBADEN   = OFF       // Port B A/D Enabler
    #pragma config LPT1OSC  = OFF       // Low power timer1 osc Enabler
    #pragma config MCLRE    = ON
    #pragma config STVREN   = ON        // Stack over/under reset enabler
    #pragma config LVP      = OFF
//  #pragma config ICPRT    = OFF       // Dedicated In-Circuit Debug/Programming
    #pragma config XINST    = OFF       // Extended Instruction Set
    #pragma config DEBUG    = OFF
//  #pragma config CP0      = OFF
//  #pragma config CP1      = OFF
//  #pragma config CP2      = OFF
//  #pragma config CP3      = OFF
//  #pragma config CPB      = OFF
//  #pragma config CPD      = OFF
//  #pragma config WRT0     = OFF
//  #pragma config WRT1     = OFF
//  #pragma config WRT2     = OFF
//  #pragma config WRT3     = OFF
//  #pragma config WRTB     = OFF       // Boot Block Write Protection
//  #pragma config WRTC     = OFF
//  #pragma config WRTD     = OFF
//  #pragma config EBTR0    = OFF
//  #pragma config EBTR1    = OFF
//  #pragma config EBTR2    = OFF
//  #pragma config EBTR3    = OFF
//  #pragma config EBTRB    = OFF

#define SYS_FREQ        16000000L       // 96MHz /6 (OSC4_PLL6)
#define FCY             SYS_FREQ/4
#define TMR0CLK         (FCY/8)         // 500kHz
#define MSEC2TMR0(x)    (TMR0CLK/1000*(x))
#define TMR1CLK         (FCY)           // 4MHz
#define MSEC2TMR1(x)    (TMR1CLK/1000*(x))      // 

/******************************************************************************/
/* Serial interface                                                           */
/******************************************************************************/

// Serial ring buffer
#define SER_BUFF_SIZE 64
unsigned char ser_buff[SER_BUFF_SIZE];
unsigned char ser_buff_ridx;
unsigned char ser_buff_widx;
unsigned char ser_buff_flags;
#define SER_BUFF_FULL        (0x01)
#define SER_BUFF_EMPTY       (0x02)
#define SER_BUFF_OVERFLOW    (0x04)

#define ser_has_data() (~ser_buff_flags & SER_BUFF_EMPTY)

void ser_init(void)
{
    // Setup bps
    SPBRGH = 0;
    //SPBRG = 51;       // 9600 bps @ 8MHz
    SPBRG = 103;        // 9600bps @ 16MHz

    TRISCbits.RC6 = 1;
    TRISCbits.RC7 = 1;
    TXSTA = 0b00100000;     // 8bit, Tx ena, Async, BRGH=0
    RCSTA = 0b10010000;     // Ser ena, 8bit, Cont rx ena
    BAUDCON = 0b00001000;   // BRG16 ena

    // Init serial ring buffer
    ser_buff_ridx = 0;
    ser_buff_widx = 0;
    ser_buff_flags = SER_BUFF_EMPTY;

    // Enable serial interrupt
    PIE1bits.RCIE = 1;
}

unsigned char ser_getc(void)
{
    unsigned char c;
    PIE1bits.RCIE = 0;
    c = ser_buff[ser_buff_ridx];
    ser_buff_flags &= ~SER_BUFF_FULL;
    if(ser_buff_ridx == ser_buff_widx -1 || ser_buff_ridx == ser_buff_widx + SER_BUFF_SIZE -1){
        // Buffer is now empty state.
        ser_buff_flags |= SER_BUFF_EMPTY;
    }
    if(ser_buff_ridx == SER_BUFF_SIZE -1){
        ser_buff_ridx = 0;
    } else {
        ser_buff_ridx ++;
    }
    PIE1bits.RCIE = 1;
    return c;
}

/******************************************************************************/
/* Econonavit data parser                                                     */
/******************************************************************************/

// Data structure
struct navitdata_st {
    unsigned long gen_int;      // Integrated generated power [kWh]
    unsigned long buy_int;      // Integrated bought power [kWh]
    unsigned long sell_int;     // Integrated sold power [kWh]
    short gen;                  // Current generated power [10W]
    short flow;                 // Current flow power [10W]
    unsigned char hour;         // Time hour
    unsigned char min;          // Time min
    unsigned char stat;         // PCU status
} ntmp;

unsigned char index = 0;
unsigned char check = 0;

struct navitdata_st * navit_retr(void)
{
    unsigned char c;
    struct navitdata_st *ndp = NULL;

    while( ser_has_data() ){
        c = ser_getc();
        if(index != 0x4e){
            check += c;
        }
        switch(index++){

        // 0x00 - 0x03 : Header
        case 0:
            if(c != 0x7e) {index = 0;} break;
        case 1:
            if(c != 0x00) {index = 0;} break;
        case 2:
            if(c != 0x4b) {index = 0;} else {check = 0;} break;
        case 3:
            if(c != 0x90) {index = 0;} break;

        // 0x2b : hour
        case 0x2b:
            ntmp.hour = c^0xa1; break;
        // 0x2c : min
        case 0x2c:
            ntmp.min = c^0xae; break;

        // 0x30 - 0x31 : current flow power
        case 0x30:
            ntmp.flow = (short)(c^0x69); break;
        case 0x31:
            ntmp.flow |= ((short)(c^0x61)) << 8; break;

        // 0x32 - 0x35 : sold int power
        case 0x32:
            ntmp.sell_int = (unsigned long)(c^0xde); break;
        case 0x33:
            ntmp.sell_int |= ((unsigned long)(c^0xf7)) << 8; break;
        case 0x34:
            ntmp.sell_int |= ((unsigned long)(c^0xc5)) << 16; break;
        case 0x35:
            ntmp.sell_int |= ((unsigned long)(c^0x4b)) << 24; break;

        // 0x36 - 0x39 : bought int power
        case 0x36:
            ntmp.buy_int = (unsigned long)(c^0x2c); break;
        case 0x37:
            ntmp.buy_int |= ((unsigned long)(c^0x9e)) << 8; break;
        case 0x38:
            ntmp.buy_int |= ((unsigned long)(c^0x82)) << 16; break;
        case 0x39:
            ntmp.buy_int |= ((unsigned long)(c^0x99)) << 24; break;

        // 0x3d - 0x3e : current PV generated power
        case 0x3d:
            ntmp.gen = (short)(c^0xd2); break;
        case 0x3e:
            ntmp.gen |= ((short)(c^0x73)) << 8; break;

        // 0x3f - 0x42 : generated int power
        case 0x3f:
            ntmp.gen_int = (unsigned long)(c^0x68); break;
        case 0x40:
            ntmp.gen_int |= ((unsigned long)(c^0x9c)) << 8; break;
        case 0x41:
            ntmp.gen_int |= ((unsigned long)(c^0x46)) << 16; break;
        case 0x42:
            ntmp.gen_int |= ((unsigned long)(c^0x96)) << 24; break;

        // 0x43 : stat
        case 0x43:
            ntmp.stat = c^0xa6; break;

        // 0x4e : Checksum
        case 0x4e:
            if(c + check == 0xff){
                // Sum OK
                ndp = &ntmp;
            }
            index = 0;
            break;

        default:
            break;

        } // switch(index++)

    } // while( ser_has_data() )

    return ndp;
}

/******************************************************************************/
/* Dynamic LED display                                                        */
/******************************************************************************/
// display ram
unsigned short disp_1, disp_2, disp_3, disp_4, disp_5, disp_6;

// Disp control
#define DISP_ON    (0x8000)
#define DISP_DP    (0x0002)

// Disp common port
#define COM_DISP1  LATAbits.LA4
#define COM_DISP2  LATAbits.LA3
#define COM_DISP3  LATAbits.LA1
#define COM_DISP4  LATAbits.LA2
#define COM_DISP5  LATAbits.LA5
#define COM_DISP6  LATAbits.LA0

// Disp duty
#define TMR1_BLANKING   (0xffff - (unsigned short)(MSEC2TMR1(0.1)))
#define TMR1_ON7G       (0xffff - (unsigned short)(MSEC2TMR1(3)))
#define TMR1_ON7R       (0xffff - (unsigned short)(MSEC2TMR1(2)))
#define TMR1_ON10G      (0xffff - (unsigned short)(MSEC2TMR1(3.3)))
#define TMR1_ON10R      (0xffff - (unsigned short)(MSEC2TMR1(1.2)))

void disp_init(void)
{
    // Setup TMR1 (dynamic scan timer)
    T1CONbits.RD16 = 1;     // Use 16bit
    T1CONbits.T1CKPS = 0;   // No prescale
    T1CONbits.TMR1CS = 0;   // Use int clock
    T1CONbits.TMR1ON = 1;
    PIE1bits.TMR1IE = 1;
}

// Segment mapping
// 7seg 10bar Port
//   a    0   RB7
//   b    1   RB5
//   c    2   RB3
//   d    3   RB0
//   e    4   RB2
//   f    5   RB6
//   g    6   RB4
//   dp   7   RB1
//        8   RC0
//        9   RC1

const unsigned short segtbl_7seg[10] = {
    //afbgcepd
    0b11101101,     // 0
    0b00101000,     // 1
    0b10110101,     // 2
    0b10111001,     // 3
    0b01111000,     // 4
    0b11011001,     // 5
    0b11011101,     // 6
    0b11101000,     // 7
    0b11111101,     // 8
    0b11111001      // 9
};

const unsigned short segtbl_10bar[11] = {
    //0194837526
    0b0000000000,   // blank
    0b1000000000,   // 1
    0b1100000000,   // 2
    0b1100000010,   // 3
    0b1100010010,   // 4
    0b1101010010,   // 5
    0b1101010110,   // 6
    0b1101010111,   // 7
    0b1101011111,   // 8
    0b1101111111,   // 9
    0b1111111111    // 10
};

const unsigned short segtbl_10barinv[11] = {
    //0194837526
    0b0000000000,   // blank
    0b0010000000,   // 1
    0b0010100000,   // 2
    0b0010101000,   // 3
    0b0010101001,   // 4
    0b0010101101,   // 5
    0b0011101101,   // 6
    0b0011111101,   // 7
    0b0011111111,   // 8
    0b0111111111,   // 9
    0b1111111111    // 10
};

void disp_update(struct navitdata_st *ndp)
{
    // Round down 10W order to make consistent with Econonavit display.
    short cons = ndp->gen + ndp->flow /10 *10;
    if(cons < 0) cons = 0;

    // Current PV generated power
    if(ndp->stat == 0x01){
        disp_1 = segtbl_7seg[MIN(ndp->gen/100, 9)];
        disp_1 |= (DISP_ON | DISP_DP);
        disp_2 = segtbl_7seg[ndp->gen%100/10];
        disp_2 |= DISP_ON;
        disp_5 = segtbl_10barinv[MIN(ndp->gen/30, 10)];
        disp_5 |= DISP_ON;
    } else {
        // Blanking when PCU is in sleep
        disp_1 &= ~DISP_ON;
        disp_2 &= ~DISP_ON;
        disp_5 &= ~DISP_ON;
    }

    // Current consumption power
    disp_3 = segtbl_7seg[MIN(cons/100, 9)];
    disp_3 |= (DISP_ON | DISP_DP);
    disp_4 = segtbl_7seg[cons%100/10];
    disp_4 |= DISP_ON;
    disp_6 = segtbl_10bar[MIN(cons/30, 10)];
    disp_6 |= DISP_ON;
}


/******************************************************************************/
/* USB Talker                                                                 */
/******************************************************************************/
void talk_to_host(struct navitdata_st *ndp)
{
    char buff[96];
    sprintf(buff, "time=%02d:%02d,stat=%d,gen=%d,flow=%d,genint=%ld,sellint=%ld,buyint=%ld\r\n",
        ndp->hour, ndp->min, ndp->stat, ndp->gen, ndp->flow, ndp->gen_int, ndp->sell_int, ndp->buy_int);
    putsUSBUSART(buff);
}

/******************************************************************************/
/* Interrupt handler                                                          */
/******************************************************************************/
unsigned char xtsk_25ms;
unsigned char disp_phase;
void interrupt isr(void)
{
    unsigned char x;
    if(INTCONbits.TMR0IF){
        // 25ms periodical timer interrupt
        TMR0 = 0xffff - MSEC2TMR0(25);
        xtsk_25ms = 1;
        INTCONbits.TMR0IF = 0;
    }

    if(PIR1bits.RCIF){
        // Serial data recv interrupt
        x = RCREG;
        if(ser_buff_flags & SER_BUFF_FULL){
            ser_buff_flags |= SER_BUFF_OVERFLOW;
        } else {
            ser_buff[ser_buff_widx] = x;
            ser_buff_flags &= ~SER_BUFF_EMPTY;
            if( ser_buff_widx == ser_buff_ridx -1 || ser_buff_widx == ser_buff_ridx + SER_BUFF_SIZE -1){
                // Buffer is now full state.
                ser_buff_flags |= SER_BUFF_FULL;
            }
            if(ser_buff_widx == SER_BUFF_SIZE -1){
                ser_buff_widx = 0;
            } else {
                ser_buff_widx ++;
            }
        }
        PIR1bits.RCIF = 0;
    }

    if(PIR1bits.TMR1IF){
        // Dynamic display interrupt
        switch(disp_phase++){
            case 0:     // Gen pow 10's
                TMR1 = TMR1_ON7G;
                LATB = (unsigned char)(disp_1 & 0xff);
                LATC = 0;
                COM_DISP1 = (disp_1 & DISP_ON) ? 1:0;
                break;

            case 2:     // Gen pow 1's
                TMR1 = TMR1_ON7G;
                LATB = (unsigned char)(disp_2 & 0xff);
                COM_DISP2 = (disp_2 & DISP_ON) ? 1:0;
                break;

            case 4:     // Cons pow 10's
                TMR1 = TMR1_ON7R;
                LATB = (unsigned char)(disp_3 & 0xff);
                COM_DISP3 = (disp_3 & DISP_ON) ? 1:0;
                break;

            case 6:     // Cons pow 1's
                TMR1 = TMR1_ON7R;
                LATB = (unsigned char)(disp_4 & 0xff);
                COM_DISP4 = (disp_4 & DISP_ON) ? 1:0;
                break;

            case 8:     // Gen pow 10bar
                TMR1 = TMR1_ON10G;
                LATB = (unsigned char)(disp_5 & 0xff);
                LATC = (unsigned char)((disp_5 >> 8) & 0xff);
                COM_DISP5 =  (disp_5 & DISP_ON) ? 1:0;
                break;

            case 10:    // Cons pow 10bar
                TMR1 = TMR1_ON10R;
                LATB = (unsigned char)(disp_6 & 0xff);
                LATC = (unsigned char)((disp_6 >> 8) & 0xff);
                COM_DISP6 =  (disp_6 & DISP_ON) ? 1:0;
                break;

            case 11:    // Blanking
                disp_phase = 0;
            case 1: case 3: case 5: case 7: case 9:
                TMR1 = TMR1_BLANKING;
                LATA = 0;
                break;

            default:
                disp_phase=0;
                break;
        } // switch(nphase++)

        PIR1bits.TMR1IF = 0;
    }
}

/******************************************************************************/
/* Main                                                                       */
/******************************************************************************/
void main(void)
{
    struct navitdata_st *ndp;

    // Setup port
    ADCON1 = 0x0F;          // Set all pins to digital
    TRISA = 0b11000000;     // RA0 - RA5 as output (as disp common)
    TRISB = 0b00000000;     // RB0 - RB7 as output (as disp segment)
    TRISC = 0b11111100;     // RC0 - RC1 as output (as disp segment)

    // Init port
    LATA = 0;
    LATB = 0;
    LATC = 0;

    // Setup task timer (TMR0)
    T0CONbits.T08BIT = 0;   // Select 16bit
    T0CONbits.T0CS = 0;     // Select internal clock
    T0CONbits.T0PS = 2;     // Select 1/8 prescale
    T0CONbits.PSA = 0;      // Select prescaler
    TMR0 = 0xFFFF;
    INTCONbits.TMR0IE = 1;
    T0CONbits.TMR0ON = 1;

    //OSCCON = 0b01110010;

    // Init USB
    USBDeviceInit();        // usb_device.c

    // Init serial
    ser_init();

    // Init display
    disp_init();

    // Enable interrupt
    INTCONbits.PEIE = 1;
    INTCONbits.GIE = 1;

    // Main task loop
    while(1){
        USBDeviceTasks();
        CDCTxService();
        if(xtsk_25ms){
            ndp = navit_retr();
            if(ndp){
                disp_update(ndp);
                talk_to_host(ndp);
            }
            xtsk_25ms = 0;
        }
    }
}
// EOF
